@@ -3,7 +3,10 @@
 
 """Miscellaneous stuff for coverage.py."""
 
+from __future__ import annotations
+
 import contextlib
+import datetime
 import errno
 import hashlib
 import importlib
@@ -12,24 +15,29 @@ import inspect
 import locale
 import os
 import os.path
-import random
 import re
-import socket
 import sys
 import types
 
+from types import ModuleType
+from typing import (
+    Any, Callable, Dict, IO, Iterable, Iterator, List, Mapping, NoReturn, Optional,
+    Sequence, Tuple, TypeVar, Union,
+)
+
 from coverage import env
 from coverage.exceptions import CoverageException
+from coverage.types import TArc
 
 # In 6.0, the exceptions moved from misc.py to exceptions.py.  But a number of
 # other packages were importing the exceptions from misc, so import them here.
 # pylint: disable=unused-wildcard-import
 from coverage.exceptions import *   # pylint: disable=wildcard-import
 
-ISOLATED_MODULES = {}
+ISOLATED_MODULES: Dict[ModuleType, ModuleType] = {}
 
 
-def isolate_module(mod):
+def isolate_module(mod: ModuleType) -> ModuleType:
     """Copy a module so that we are isolated from aggressive mocking.
 
     If a test suite mocks os.path.exists (for example), and then we need to use
@@ -52,10 +60,10 @@ os = isolate_module(os)
 
 class SysModuleSaver:
     """Saves the contents of sys.modules, and removes new modules later."""
-    def __init__(self):
+    def __init__(self) -> None:
         self.old_modules = set(sys.modules)
 
-    def restore(self):
+    def restore(self) -> None:
         """Remove any modules imported since this object started."""
         new_modules = set(sys.modules) - self.old_modules
         for m in new_modules:
@@ -63,7 +71,7 @@ class SysModuleSaver:
 
 
 @contextlib.contextmanager
-def sys_modules_saved():
+def sys_modules_saved() -> Iterator[None]:
     """A context manager to remove any modules imported during a block."""
     saver = SysModuleSaver()
     try:
@@ -72,7 +80,7 @@ def sys_modules_saved():
         saver.restore()
 
 
-def import_third_party(modname):
+def import_third_party(modname: str) -> Tuple[ModuleType, bool]:
     """Import a third-party module we need, but might not be installed.
 
     This also cleans out the module after the import, so that coverage won't
@@ -83,69 +91,19 @@ def import_third_party(modname):
         modname (str): the name of the module to import.
 
     Returns:
-        The imported module, or None if the module couldn't be imported.
+        The imported module, and a boolean indicating if the module could be imported.
+
+    If the boolean is False, the module returned is not the one you want: don't use it.
 
     """
     with sys_modules_saved():
         try:
-            return importlib.import_module(modname)
+            return importlib.import_module(modname), True
         except ImportError:
-            return None
+            return sys, False
 
 
-def dummy_decorator_with_args(*args_unused, **kwargs_unused):
-    """Dummy no-op implementation of a decorator with arguments."""
-    def _decorator(func):
-        return func
-    return _decorator
-
-
-# Environment COVERAGE_NO_CONTRACTS=1 can turn off contracts while debugging
-# tests to remove noise from stack traces.
-# $set_env.py: COVERAGE_NO_CONTRACTS - Disable PyContracts to simplify stack traces.
-USE_CONTRACTS = env.TESTING and not bool(int(os.environ.get("COVERAGE_NO_CONTRACTS", 0)))
-
-# Use PyContracts for assertion testing on parameters and returns, but only if
-# we are running our own test suite.
-if USE_CONTRACTS:
-    from contracts import contract              # pylint: disable=unused-import
-    from contracts import new_contract as raw_new_contract
-
-    def new_contract(*args, **kwargs):
-        """A proxy for contracts.new_contract that doesn't mind happening twice."""
-        try:
-            raw_new_contract(*args, **kwargs)
-        except ValueError:
-            # During meta-coverage, this module is imported twice, and
-            # PyContracts doesn't like redefining contracts. It's OK.
-            pass
-
-    # Define contract words that PyContract doesn't have.
-    new_contract('bytes', lambda v: isinstance(v, bytes))
-    new_contract('unicode', lambda v: isinstance(v, str))
-
-    def one_of(argnames):
-        """Ensure that only one of the argnames is non-None."""
-        def _decorator(func):
-            argnameset = {name.strip() for name in argnames.split(",")}
-            def _wrapper(*args, **kwargs):
-                vals = [kwargs.get(name) for name in argnameset]
-                assert sum(val is not None for val in vals) == 1
-                return func(*args, **kwargs)
-            return _wrapper
-        return _decorator
-else:                                           # pragma: not testing
-    # We aren't using real PyContracts, so just define our decorators as
-    # stunt-double no-ops.
-    contract = dummy_decorator_with_args
-    one_of = dummy_decorator_with_args
-
-    def new_contract(*args_unused, **kwargs_unused):
-        """Dummy no-op implementation of `new_contract`."""
-        pass
-
-
-def nice_pair(pair):
+def nice_pair(pair: TArc) -> str:
     """Make a nice string representation of a pair of numbers.
 
     If the numbers are equal, just return the number, otherwise return the pair
@@ -159,7 +117,10 @@ def nice_pair(pair):
         return "%d-%d" % (start, end)
 
 
-def expensive(fn):
+TSelf = TypeVar("TSelf")
+TRetVal = TypeVar("TRetVal")
+
+def expensive(fn: Callable[[TSelf], TRetVal]) -> Callable[[TSelf], TRetVal]:
     """A decorator to indicate that a method shouldn't be called more than once.
 
     Normally, this does nothing.  During testing, this raises an exception if
@@ -169,7 +130,7 @@ def expensive(fn):
     if env.TESTING:
         attr = "_once_" + fn.__name__
 
-        def _wrapper(self):
+        def _wrapper(self: TSelf) -> TRetVal:
             if hasattr(self, attr):
                 raise AssertionError(f"Shouldn't have called {fn.__name__} more than once")
             setattr(self, attr, True)
@@ -179,7 +140,7 @@ def expensive(fn):
         return fn                   # pragma: not testing
 
 
-def bool_or_none(b):
+def bool_or_none(b: Any) -> Optional[bool]:
     """Return bool(b), but preserve None."""
     if b is None:
         return None
@@ -187,12 +148,16 @@ def bool_or_none(b):
         return bool(b)
 
 
-def join_regex(regexes):
-    """Combine a list of regexes into one that matches any of them."""
-    return "|".join("(?:%s)" % r for r in regexes)
+def join_regex(regexes: Iterable[str]) -> str:
+    """Combine a series of regex strings into one that matches any of them."""
+    regexes = list(regexes)
+    if len(regexes) == 1:
+        return regexes[0]
+    else:
+        return "|".join(f"(?:{r})" for r in regexes)
 
 
-def file_be_gone(path):
+def file_be_gone(path: str) -> None:
     """Remove a file, and don't get annoyed if it doesn't exist."""
     try:
         os.remove(path)
@@ -201,7 +166,7 @@ def file_be_gone(path):
             raise
 
 
-def ensure_dir(directory):
+def ensure_dir(directory: str) -> None:
     """Make sure the directory exists.
 
     If `directory` is None or empty, do nothing.
@@ -210,12 +175,12 @@ def ensure_dir(directory):
         os.makedirs(directory, exist_ok=True)
 
 
-def ensure_dir_for_file(path):
+def ensure_dir_for_file(path: str) -> None:
     """Make sure the directory for the path exists."""
     ensure_dir(os.path.dirname(path))
 
 
-def output_encoding(outfile=None):
+def output_encoding(outfile: Optional[IO[str]] = None) -> str:
     """Determine the encoding to use for output written to `outfile` or stdout."""
     if outfile is None:
         outfile = sys.stdout
@@ -227,32 +192,12 @@ def output_encoding(outfile=None):
     return encoding
 
 
-def filename_suffix(suffix):
-    """Compute a filename suffix for a data file.
-
-    If `suffix` is a string or None, simply return it. If `suffix` is True,
-    then build a suffix incorporating the hostname, process id, and a random
-    number.
-
-    Returns a string or None.
-
-    """
-    if suffix is True:
-        # If data_suffix was a simple true value, then make a suffix with
-        # plenty of distinguishing information.  We do this here in
-        # `save()` at the last minute so that the pid will be correct even
-        # if the process forks.
-        dice = random.Random(os.urandom(8)).randint(0, 999999)
-        suffix = "%s.%s.%06d" % (socket.gethostname(), os.getpid(), dice)
-    return suffix
-
-
 class Hasher:
     """Hashes Python data for fingerprinting."""
-    def __init__(self):
+    def __init__(self) -> None:
         self.hash = hashlib.new("sha3_256")
 
-    def update(self, v):
+    def update(self, v: Any) -> None:
         """Add `v` to the hash, recursively if needed."""
         self.hash.update(str(type(v)).encode("utf-8"))
         if isinstance(v, str):
@@ -273,21 +218,21 @@ class Hasher:
                 self.update(v[k])
         else:
             for k in dir(v):
-                if k.startswith('__'):
+                if k.startswith("__"):
                     continue
                 a = getattr(v, k)
                 if inspect.isroutine(a):
                     continue
                 self.update(k)
                 self.update(a)
-        self.hash.update(b'.')
+        self.hash.update(b".")
 
-    def hexdigest(self):
+    def hexdigest(self) -> str:
         """Retrieve the hex digest of the hash."""
         return self.hash.hexdigest()[:32]
 
 
-def _needs_to_implement(that, func_name):
+def _needs_to_implement(that: Any, func_name: str) -> NoReturn:
     """Helper to raise NotImplementedError in interface stubs."""
     if hasattr(that, "_coverage_plugin_name"):
         thing = "Plugin"
@@ -299,7 +244,7 @@ def _needs_to_implement(that, func_name):
 
     raise NotImplementedError(
         f"{thing} {name!r} needs to implement {func_name}()"
-        )
+    )
 
 
 class DefaultValue:
@@ -309,14 +254,14 @@ class DefaultValue:
     and Sphinx output.
 
     """
-    def __init__(self, display_as):
+    def __init__(self, display_as: str) -> None:
         self.display_as = display_as
 
-    def __repr__(self):
+    def __repr__(self) -> str:
         return self.display_as
 
 
-def substitute_variables(text, variables):
+def substitute_variables(text: str, variables: Mapping[str, str]) -> str:
     """Substitute ``${VAR}`` variables in `text` with their values.
 
     Variables in the text can take a number of shell-inspired forms::
@@ -347,33 +292,33 @@ def substitute_variables(text, variables):
         )
         """
 
-    dollar_groups = ('dollar', 'word1', 'word2')
+    dollar_groups = ("dollar", "word1", "word2")
 
-    def dollar_replace(match):
+    def dollar_replace(match: re.Match[str]) -> str:
         """Called for each $replacement."""
-        # Only one of the groups will have matched, just get its text.
+        # Only one of the dollar_groups will have matched, just get its text.
         word = next(g for g in match.group(*dollar_groups) if g)    # pragma: always breaks
         if word == "$":
             return "$"
         elif word in variables:
             return variables[word]
-        elif match.group('strict'):
+        elif match["strict"]:
             msg = f"Variable {word} is undefined: {text!r}"
             raise CoverageException(msg)
         else:
-            return match.group('defval')
+            return match["defval"]
 
     text = re.sub(dollar_pattern, dollar_replace, text)
     return text
 
 
-def format_local_datetime(dt):
+def format_local_datetime(dt: datetime.datetime) -> str:
     """Return a string with local timezone representing the date.
     """
-    return dt.astimezone().strftime('%Y-%m-%d %H:%M %z')
+    return dt.astimezone().strftime("%Y-%m-%d %H:%M %z")
 
 
-def import_local_file(modname, modfile=None):
+def import_local_file(modname: str, modfile: Optional[str] = None) -> ModuleType:
     """Import a local file as a module.
 
     Opens a file in the current directory named `modname`.py, imports it
@@ -382,29 +327,35 @@ def import_local_file(modname, modfile=None):
 
     """
     if modfile is None:
-        modfile = modname + '.py'
+        modfile = modname + ".py"
     spec = importlib.util.spec_from_file_location(modname, modfile)
+    assert spec is not None
     mod = importlib.util.module_from_spec(spec)
     sys.modules[modname] = mod
+    assert spec.loader is not None
     spec.loader.exec_module(mod)
 
     return mod
 
 
-def human_key(s):
+def _human_key(s: str) -> Tuple[List[Union[str, int]], str]:
     """Turn a string into a list of string and number chunks.
-        "z23a" -> ["z", 23, "a"]
+
+    "z23a" -> (["z", 23, "a"], "z23a")
+
+    The original string is appended as a last value to ensure the
+    key is unique enough so that "x1y" and "x001y" can be distinguished.
     """
-    def tryint(s):
+    def tryint(s: str) -> Union[str, int]:
         """If `s` is a number, return an int, else `s` unchanged."""
         try:
             return int(s)
         except ValueError:
             return s
 
-    return [tryint(c) for c in re.split(r"(\d+)", s)]
+    return ([tryint(c) for c in re.split(r"(\d+)", s)], s)
 
-def human_sorted(strings):
+def human_sorted(strings: Iterable[str]) -> List[str]:
     """Sort the given iterable of strings the way that humans expect.
 
     Numeric components in the strings are sorted as numbers.
@@ -412,11 +363,42 @@ def human_sorted(strings):
     Returns the sorted list.
 
     """
-    return sorted(strings, key=human_key)
+    return sorted(strings, key=_human_key)
 
-def human_sorted_items(items, reverse=False):
-    """Sort the (string, value) items the way humans expect.
+SortableItem = TypeVar("SortableItem", bound=Sequence[Any])
+
+def human_sorted_items(
+    items: Iterable[SortableItem],
+    reverse: bool = False,
+) -> List[SortableItem]:
+    """Sort (string, ...) items the way humans expect.
+
+    The elements of `items` can be any tuple/list. They'll be sorted by the
+    first element (a string), with ties broken by the remaining elements.
 
     Returns the sorted list of items.
     """
-    return sorted(items, key=lambda pair: (human_key(pair[0]), pair[1]), reverse=reverse)
+    return sorted(items, key=lambda item: (_human_key(item[0]), *item[1:]), reverse=reverse)
+
+
+def plural(n: int, thing: str = "", things: str = "") -> str:
+    """Pluralize a word.
+
+    If n is 1, return thing.  Otherwise return things, or thing+s.
+    """
+    if n == 1:
+        return thing
+    else:
+        return things or (thing + "s")
+
+
+def stdout_link(text: str, url: str) -> str:
+    """Format text+url as a clickable link for stdout.
+
+    If attached to a terminal, use escape sequences. Otherwise, just return
+    the text.
+    """
+    if hasattr(sys.stdout, "isatty") and sys.stdout.isatty():
+        return f"\033]8;;{url}\a{text}\033]8;;\a"
+    else:
+        return text
